/*
   Soubor: proxyserv.cc
   Autor:  Pavel Lipensky (xlipen@stud.fit.vutbr.cz)
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>

#include <iostream>
#include <fstream>
#include <cstring>
#include <cstdio>

using namespace std;

#define MAX_CLIENTS 5
#define BUFFER_LEN  4096

int mode, head, cnt;

void errmsg(char *msg)
{
   if (*msg)  cerr << "PROXYSERV: " << msg << "\n";
   else cerr << "Usage: poxyserv -tcp|udp serverport clientport [-h]\n\n";
   exit(1);
}


void tcp_write(ostream &os, const char* buffer, int len, int *hd) { // zapis dat
   const char *pend;
   int l;
   if (head && *hd &&    // pozadovana jen hlavicka && priznak halvicky v bloku
       ( (l=4, pend=strstr(buffer, "\x0D\x0A\x0D\x0A")) ||  // M$ format 
         (l=2, pend=strstr(buffer, "\x0A\x0A")) )   ) {     // UNIX format
      os.write(buffer, pend-buffer+l); // zapis jen hlavicku
      *hd=0; // hlavicka odfiltrovana, dalsi bloky se nefiltruji a nezapisuji
   }
   if (*hd) os.write(buffer, len);
}


void tcp_transfer (int c_sck, int s_sck) // zpracovani pozadavku 
{
   // buffer se zarazkou pro strstr() neuspesne hledani konce hlavicky
   char buffer[BUFFER_LEN+1] = {0}; 
   int c_len=BUFFER_LEN, s_len=BUFFER_LEN,  // aby se hned nezavrel soket
       c_head=1, s_head=1;
   
   // zajisteni neblokujiciho cteni
   struct timeval wait;
   fd_set tmp;
   int nb;

   // vytvoreni vystupniho streamu pro kazdy prenos
   char c_name[10], s_name[10]; 
   sprintf(c_name, "sck%03i.clt", cnt%1000);
   sprintf(s_name, "sck%03i.srv", cnt%1000);
   ofstream c_os(c_name, ios::out | ios::binary); 
   ofstream s_os(s_name, ios::out | ios::binary); 
   
   do {
      wait.tv_sec = 1;
      wait.tv_usec =0;
      FD_ZERO(&tmp);
      FD_SET(c_sck,&tmp);
      FD_SET(s_sck,&tmp);

      // vyber soketu ktery muze pouzit (blokujici) read() 
      nb=select(FD_SETSIZE,&tmp,(fd_set*)NULL,(fd_set*)NULL,&wait);
      if (nb<0) errmsg("select() failed or time out");
   
      if(FD_ISSET(c_sck, &tmp)) { // cteni dat od klienta
         
         if ((c_len = read(c_sck, buffer, BUFFER_LEN)) < 0)
            errmsg("Bad recieve by client");
         if (c_len > 0){ // ulozeni a zaslani dat serveru
            tcp_write(c_os, buffer, c_len, &c_head);
            if (write(s_sck, buffer, c_len) < 0) 
               errmsg("Cannot send data to server");
         }
//         if (c_len < BUFFER_LEN) c_head=1; // obnoveni filtrovani !!!                       
      }

      if(FD_ISSET(s_sck, &tmp)) { // cteni dat od serveru

         if ((s_len = read(s_sck, buffer, BUFFER_LEN)) < 0) 
            errmsg("Bad recieve by server");        
         if (s_len > 0) { // ulozeni a zaslani dat klientovi       
            tcp_write(s_os, buffer, s_len, &s_head); // zapis do souboru  
            if (write(c_sck, buffer, s_len) < 0)     
               errmsg("Cannot send data to client");
         }
//         if (s_len < BUFFER_LEN) s_head=1; // obnoveni filtrovani !!!                       
      }
  
   // dokud neskonci prenos jedne zpravy tam a zpet 
   } while (c_len==BUFFER_LEN || s_len==BUFFER_LEN);
/*
   // efektivni - vice zprav v soketu,  nelze filtrovat vsechny hlavicky !!!
   } while (c_len > 0 || s_len > 0); // chteji spolu dal komunikovat?
*/
}


int m_sck, c_sck, s_sck;
struct sockaddr_in addr0, addr1, addr2;


void tcp_proxyserv(const char *s_port, const char *c_port) 
{
   // vytvoreni hlavniho socketu
   if ( (m_sck = socket(AF_INET, mode, 0)) < 0) 
      errmsg("Cannot open main socket");

   // vytvoreni datove sruktury pro ulozeni adresy a portu
   addr0.sin_family = AF_INET;
   addr0.sin_addr.s_addr = htonl(INADDR_ANY);
   addr0.sin_port = htons(atoi(s_port)); // zdrojovy port

   // prideleni a overeni skutecne prideleneho portu
   socklen_t len = sizeof(addr0);
   if (bind (m_sck, (struct sockaddr*) &addr0, len))
      errmsg("bind() failed");
   if (getsockname(m_sck, (struct sockaddr*) &addr0, &len))
      errmsg("getsockname() failed");
      
   cerr << "PROXYSERV: Port number is " << ntohs(addr0.sin_port) << "\n";
   listen(m_sck, MAX_CLIENTS);
   signal(SIGCHLD,SIG_IGN);

   for (;;) {  // cyklus obsluhy pozadavku
   
      socklen_t len = sizeof(addr0);   
      if ((c_sck=accept(m_sck, (struct sockaddr*) &addr0, &len)) < 0)  // ceka na pozadavek
         errmsg("Bad client socket");
      else {
         pid_t pid;
         if ( (pid = fork()) < 0) 
            errmsg("Failed to create child process");
         else if (pid == 0) { // syn zpracuje pozadavek a skonci
            close(m_sck);     // nestara se o otce

            if ((s_sck = socket(AF_INET, mode, 0)) < 0)
               errmsg("Cannot open server socket");
               
            addr1.sin_family = AF_INET;   
            addr1.sin_addr.s_addr = htonl(INADDR_ANY);
            addr1.sin_port = htons(atoi(c_port));  // cilovy port

           // pripojeni k serveru
           if (connect(s_sck, (struct sockaddr*) &addr1, sizeof(addr1)) < 0) 
               errmsg("Cannot connect server");
        
            tcp_transfer(c_sck, s_sck); // vymena dat pres tcp

            close(s_sck);      
            close(c_sck);
            return;               // syn konci
         }
         else {                     // otec (nestara se o syna, jen o pozadavky)
	    close(c_sck);
	    cnt++;                  // pocitadlo pozadavku
	 }
      }
   } // for     
}


void udp_proxyserv(const char *s_port, const char *c_port) 
{
   // nataveni serveru
   addr0.sin_family = AF_INET;
   addr0.sin_addr.s_addr = INADDR_ANY;
   addr0.sin_port = htons(atoi(s_port));
   unsigned int addr0_len = sizeof(addr0);
    
   // napojeni datagramoveho serveru
   if ((s_sck = socket(AF_INET, mode, 0)) < 0)
      errmsg("Cannot open server socket");
   if (bind(s_sck,(struct sockaddr *)&addr0, addr0_len) < 0)
      errmsg("bind() failed (server)");        

   // nataveni klienta
   addr1.sin_family = AF_INET;
   addr1.sin_port = htons(atoi(c_port));
   addr1.sin_addr.s_addr = INADDR_ANY;
   unsigned int addr1_len = sizeof(addr1);
   
   addr2.sin_family = AF_INET;
   addr2.sin_port = htons(0);
   addr2.sin_addr.s_addr = INADDR_ANY;
   unsigned int addr2_len = sizeof(addr2);   
      
   if ((c_sck = socket(AF_INET, mode, 0)) < 0)
      errmsg("Cannot open client socket");
   if (bind(c_sck,(struct sockaddr *)&addr2, addr2_len) < 0)
      errmsg("bind() failed (client)");        
   
   // buffer 
   char buffer[BUFFER_LEN]; 
   int c_len, s_len; 
    
   // zajisteni neblokujiciho cteni
   struct timeval wait;
   fd_set tmp;
   int nb;

   // vytvoreni vystupniho streamu pro kazdy smer
   ofstream c_os("sckudp.srv", ios::out | ios::binary); 
   ofstream s_os("sckudp.clt", ios::out | ios::binary); 
   
   for(;;) {
      wait.tv_sec = 1;
      wait.tv_usec =0;
      FD_ZERO(&tmp);
      FD_SET(c_sck,&tmp);
      FD_SET(s_sck,&tmp);

      // vyber soketu ktery muze pouzit (blokujici) read() 
      nb=select(FD_SETSIZE,&tmp,(fd_set*)NULL,(fd_set*)NULL,&wait);
      if (nb<0) errmsg("select() failed or time out");

      if(FD_ISSET(c_sck,&tmp)) { // cteni datagramu klienta
      
         if ((c_len = recvfrom(c_sck, buffer, BUFFER_LEN, 0, 
                        (struct sockaddr *)&addr2, &addr2_len)) < 0)
            errmsg("Bad recieve by client");            
         if (c_len > 0) {
            c_os.write(buffer, c_len);
            c_os.flush(); // vynuti si zapis na disk
            if (sendto(s_sck, buffer, c_len, 0,
                        (struct sockaddr *)&addr0, addr0_len) != c_len)
               errmsg("Cannot send data to server");              
         }
      } 

      if(FD_ISSET(s_sck,&tmp)) { // cteni datagramu serveru

         if ((s_len = recvfrom(s_sck, buffer, BUFFER_LEN, 0,
                        (struct sockaddr *)&addr0, &addr0_len)) < 0)
            errmsg("Bad recieve by server");            
         if (s_len > 0) {
            s_os.write(buffer, s_len);
            s_os.flush(); // vynuti si zapis na disk
            if (sendto(c_sck, buffer, s_len, 0,
                        (struct sockaddr *)&addr1, addr1_len) != s_len)
               errmsg("Cannot send data to client");                                  
         }
      } 
   } // for 
}


int main(int argc, char *argv[])
{
   // kotrola parametru
   if(argc<4 || argc>5) errmsg("");
   if (!strcmp(argv[1], "-tcp")) mode = SOCK_STREAM; else              // TCP
   if (!strcmp(argv[1], "-udp")) mode = SOCK_DGRAM;  else errmsg("");  // UDP
   if (argc==5)  
     if (!strcmp(argv[3], "80") && !strcmp(argv[4],  "-h")) head = 1; 
     else errmsg("");
   
   if (mode==SOCK_STREAM) tcp_proxyserv(argv[2], argv[3]); 
   else udp_proxyserv(argv[2], argv[3]);
   
   return 0;
}
